from typing import Set, List

ctx_opts: Set[int]
bytes_sockopts: Set[int]
fd_sockopts: Set[int]
int_sockopts: Set[int]
int64_sockopts: Set[int]

ctx_opt_names: List[str]
bytes_sockopt_names: List[str]
fd_sockopt_names: List[str]
int_sockopt_names: List[str]
int64_sockopt_names: List[str]

DRAFT_API: bool

VERSION: int
VERSION_MAJOR: int
VERSION_MINOR: int
VERSION_PATCH: int
NOBLOCK: int
DONTWAIT: int
POLLIN: int
POLLOUT: int
POLLERR: int
POLLPRI: int
SNDMORE: int
STREAMER: int
FORWARDER: int
QUEUE: int
IO_THREADS_DFLT: int
MAX_SOCKETS_DFLT: int
POLLITEMS_DFLT: int
THREAD_PRIORITY_DFLT: int
THREAD_SCHED_POLICY_DFLT: int
PAIR: int
PUB: int
SUB: int
REQ: int
REP: int
DEALER: int
ROUTER: int
XREQ: int
XREP: int
PULL: int
PUSH: int
XPUB: int
XSUB: int
UPSTREAM: int
DOWNSTREAM: int
STREAM: int
SERVER: int
CLIENT: int
RADIO: int
DISH: int
GATHER: int
SCATTER: int
DGRAM: int
EVENT_CONNECTED: int
EVENT_CONNECT_DELAYED: int
EVENT_CONNECT_RETRIED: int
EVENT_LISTENING: int
EVENT_BIND_FAILED: int
EVENT_ACCEPTED: int
EVENT_ACCEPT_FAILED: int
EVENT_CLOSED: int
EVENT_CLOSE_FAILED: int
EVENT_DISCONNECTED: int
EVENT_ALL: int
EVENT_MONITOR_STOPPED: int
EVENT_HANDSHAKE_FAILED_NO_DETAIL: int
EVENT_HANDSHAKE_SUCCEEDED: int
EVENT_HANDSHAKE_FAILED_PROTOCOL: int
EVENT_HANDSHAKE_FAILED_AUTH: int
PROTOCOL_ERROR_ZMTP_UNSPECIFIED: int
PROTOCOL_ERROR_ZMTP_UNEXPECTED_COMMAND: int
PROTOCOL_ERROR_ZMTP_INVALID_SEQUENCE: int
PROTOCOL_ERROR_ZMTP_KEY_EXCHANGE: int
PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_UNSPECIFIED: int
PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_MESSAGE: int
PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_HELLO: int
PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_INITIATE: int
PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_ERROR: int
PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_READY: int
PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_WELCOME: int
PROTOCOL_ERROR_ZMTP_INVALID_METADATA: int
PROTOCOL_ERROR_ZMTP_CRYPTOGRAPHIC: int
PROTOCOL_ERROR_ZMTP_MECHANISM_MISMATCH: int
PROTOCOL_ERROR_ZAP_UNSPECIFIED: int
PROTOCOL_ERROR_ZAP_MALFORMED_REPLY: int
PROTOCOL_ERROR_ZAP_BAD_REQUEST_ID: int
PROTOCOL_ERROR_ZAP_BAD_VERSION: int
PROTOCOL_ERROR_ZAP_INVALID_STATUS_CODE: int
PROTOCOL_ERROR_ZAP_INVALID_METADATA: int
NOTIFY_CONNECT: int
NOTIFY_DISCONNECT: int
NULL: int
PLAIN: int
CURVE: int
GSSAPI: int
GSSAPI_NT_HOSTBASED: int
GSSAPI_NT_USER_NAME: int
GSSAPI_NT_KRB5_PRINCIPAL: int
EAGAIN: int
EINVAL: int
EFAULT: int
ENOMEM: int
ENODEV: int
EMSGSIZE: int
EAFNOSUPPORT: int
ENETUNREACH: int
ECONNABORTED: int
ECONNRESET: int
ENOTCONN: int
ETIMEDOUT: int
EHOSTUNREACH: int
ENETRESET: int
HAUSNUMERO: int
ENOTSUP: int
EPROTONOSUPPORT: int
ENOBUFS: int
ENETDOWN: int
EADDRINUSE: int
EADDRNOTAVAIL: int
ECONNREFUSED: int
EINPROGRESS: int
ENOTSOCK: int
EFSM: int
ENOCOMPATPROTO: int
ETERM: int
EMTHREAD: int
IO_THREADS: int
MAX_SOCKETS: int
SOCKET_LIMIT: int
THREAD_PRIORITY: int
THREAD_SCHED_POLICY: int
BLOCKY: int
MSG_T_SIZE: int
THREAD_AFFINITY_CPU_ADD: int
THREAD_AFFINITY_CPU_REMOVE: int
THREAD_NAME_PREFIX: int
IDENTITY: int
SUBSCRIBE: int
UNSUBSCRIBE: int
LAST_ENDPOINT: int
TCP_ACCEPT_FILTER: int
PLAIN_USERNAME: int
PLAIN_PASSWORD: int
CURVE_PUBLICKEY: int
CURVE_SECRETKEY: int
CURVE_SERVERKEY: int
ZAP_DOMAIN: int
CONNECT_RID: int
GSSAPI_PRINCIPAL: int
GSSAPI_SERVICE_PRINCIPAL: int
SOCKS_PROXY: int
XPUB_WELCOME_MSG: int
ROUTING_ID: int
CONNECT_ROUTING_ID: int
BINDTODEVICE: int
FD: int
RECONNECT_IVL_MAX: int
SNDTIMEO: int
RCVTIMEO: int
SNDHWM: int
RCVHWM: int
MULTICAST_HOPS: int
IPV4ONLY: int
ROUTER_BEHAVIOR: int
TCP_KEEPALIVE: int
TCP_KEEPALIVE_CNT: int
TCP_KEEPALIVE_IDLE: int
TCP_KEEPALIVE_INTVL: int
DELAY_ATTACH_ON_CONNECT: int
XPUB_VERBOSE: int
EVENTS: int
TYPE: int
LINGER: int
RECONNECT_IVL: int
BACKLOG: int
ROUTER_MANDATORY: int
FAIL_UNROUTABLE: int
ROUTER_RAW: int
IMMEDIATE: int
IPV6: int
MECHANISM: int
PLAIN_SERVER: int
CURVE_SERVER: int
PROBE_ROUTER: int
REQ_RELAXED: int
REQ_CORRELATE: int
CONFLATE: int
ROUTER_HANDOVER: int
TOS: int
IPC_FILTER_PID: int
IPC_FILTER_UID: int
IPC_FILTER_GID: int
GSSAPI_SERVER: int
GSSAPI_PLAINTEXT: int
HANDSHAKE_IVL: int
XPUB_NODROP: int
XPUB_MANUAL: int
STREAM_NOTIFY: int
INVERT_MATCHING: int
XPUB_VERBOSER: int
HEARTBEAT_IVL: int
HEARTBEAT_TTL: int
HEARTBEAT_TIMEOUT: int
CONNECT_TIMEOUT: int
TCP_MAXRT: int
THREAD_SAFE: int
MULTICAST_MAXTPDU: int
VMCI_CONNECT_TIMEOUT: int
USE_FD: int
GSSAPI_PRINCIPAL_NAMETYPE: int
GSSAPI_SERVICE_PRINCIPAL_NAMETYPE: int
MULTICAST_LOOP: int
ROUTER_NOTIFY: int
ZAP_ENFORCE_DOMAIN: int
AFFINITY: int
MAXMSGSIZE: int
HWM: int
SWAP: int
MCAST_LOOP: int
RECOVERY_IVL_MSEC: int
VMCI_BUFFER_SIZE: int
VMCI_BUFFER_MIN_SIZE: int
VMCI_BUFFER_MAX_SIZE: int
RATE: int
RECOVERY_IVL: int
SNDBUF: int
RCVBUF: int
RCVMORE: int
MORE: int
SRCFD: int
SHARED: int
